from enum import Enum
import functools
import logging
import time

import requests
from requests import HTTPError

from inventory_provider import environment


# Navigation Properties
# http://149.210.162.190:81/ImsVersions/4.19.9/html/86d07a57-fa45-835e-d4a2-a789c4acbc96.htm  # noqa

logger = logging.getLogger(__name__)
log_entry_and_exit = functools.partial(
    environment.log_entry_and_exit, logger=logger)


# http://149.210.162.190:81/ImsVersions/21.9/html/50e6a1b1-3910-2091-63d5-e13777b2194e.htm  # noqa
CIRCUIT_CUSTOMER_RELATION = {
    'Circuit': 2,
    'Customer': 4
}
# http://149.210.162.190:81/ImsVersions/20.1/html/86d07a57-fa45-835e-d4a2-a789c4acbc96.htm  # noqa
CIRCUIT_PROPERTIES = {
    'Site': 8,
    'Speed': 16,
    'Customer': 32,
    'Product': 128,
    'CalculatedNode': 256,
    'Ports': 512,
    'InternalPorts': 1024,
    'CarrierCircuits': 65536,
    'SubCircuits': 131072,
    'PortsFullDetails': 262144,
    'InternalPortsFullDetails': 524288,
    'PortA': 34359738368,
    'PortB': 68719476736
}
# http://149.210.162.190:81/ImsVersions/4.19.9/html/347cb410-8c05-47bd-ceb0-d1dd05bf98a4.htm  # noqa
CITY_PROPERTIES = {
    'Country': 8
}
# http://149.210.162.190:81/ImsVersions/4.19.9/html/dbc969d0-e735-132e-6281-f724c6d7da64.htm  # noqa
CONTACT_PROPERTIES = {
    'SiteRelatedContacts': 8,
    'CustomerRelatedContacts': 16,
    'GroupRelatedContacts': 32,
    'VendorRelatedContacts': 64
}
# http://149.210.162.190:81/ImsVersions/4.19.9/html/5a40472e-48ee-c120-0a36-52a85d52127c.htm  # noqa
CUSTOMER_PROPERTIES = {
    'CustomerRelatedContacts': 32768,
    'CustomerType': 262144
}
# http://149.210.162.190:81/ImsVersions/20.1/html/4f3e1d5e-53c3-1beb-cb16-65b5308dcb81.htm
CUSTOMER_RELATED_CONTACT_PROPERTIES = {
    'Customer': 8,
    'Contact': 16
}
# http://149.210.162.190:81/ImsVersions/4.19.9/html/a8dc6266-d934-8162-4a55-9e1648187f2c.htm  # noqa
EQUIP_DEF_PROPERTIES = {
    'Vendor': 8,
    'Nodes': 4096
}
# http://149.210.162.190:81/ImsVersions/20.1/html/6fd3a968-26e2-e40f-e3cd-c99afa34c3e6.htm
EXTRA_FIELD_VALUE_PROPERTIES = {
    'ExtraFieldValueObjectInfo': 64
}
# http://149.210.162.190:81/ImsVersions/4.19.9/html/f18222e3-e353-0abe-b89c-820db87940ac.htm  # noqa
INTERNAL_PORT_PROPERTIES = {
    'Node': 8,
    'Shelf': 64,
    'Circuit': 256,
}
# http://149.210.162.190:81/ImsVersions/4.19.9/html/2d27d509-77cb-537d-3ffa-796de7e82af8.htm  # noqa
NODE_PROPERTIES = {
    'EquipmentDefinition': 16,
    'ManagementSystem': 128,
    'Site': 256,
    'Shelves': 1024,
    'Ports': 4096,
    'InternalPorts': 8192,
    'NodeCity': 32768,
    'Order': 67108864,
    'NodeCountry': 1073741824
}
# http://149.210.162.190:81/ImsVersions/4.19.9/html/199f32b5-5104-fec7-8787-0d730113e902.htm  # noqa
PORT_PROPERTIES = {
    'Node': 16,
    'Shelf': 32,
    'Circuit': 512,
}
# http://149.210.162.190:81/ImsVersions/4.19.9/html/9c8d50f0-842c-5959-0fa6-14e1720669ec.htm  # noqa
SITE_PROPERTIES = {
    'City': 2,
    'SiteAliases': 64,
    'Country': 256,
    'Nodes': 32768
}
# http://149.210.162.190:81/ImsVersions/4.19.9/html/8ce06cb7-7707-46c4-f02f-86083310d81b.htm  # noqa
VENDOR_PROPERTIES = {
    'VendorRelatedContacts': 64,
    'VendorType': 1024
}
# http://149.210.162.190:81/ImsVersions/4.19.9/html/62e7aa63-aff0-6992-7697-377ace239c4f.htm  # noqa
VENDOR_RELATED_CONTACT_PROPERTIES = {
    'Vendor': 8,
    'Contact': 16
}

VM_PORT_RELATE_PROPERTIES = {
    'Circuit': 32
}

VM_INTERNAL_PORT_RELATE_PROPERTIES = {
    'Circuit': 32
}

NO_FILTERED_RESULTS_MESSAGE = 'no records found for entity:'

# this will be obsolete as soon as Inventory Provider update is done, but is
# here for between the time of the roll out and the Inventory Update
IMS_SERVICE_NAMES = {
    'EUMETSAT GRE',
    'EUMETSAT INTERNATIONAL',
    'EUMETSAT TERRESTRIAL',
    'EXPRESS ROUTE',
    'GEANT - GBS',
    'GEANT CLOUD PEERING',
    'GEANT IP',
    'GEANT LAMBDA',
    'GEANT OPEN CROSS CONNECT',
    'GEANT OPEN PORT',
    'GEANT PEERING',
    'GEANT PLUS',
    'GTS',
    'GWS - BROKERED',
    'GWS - DIRECT',
    'GWS - INDIRECT',
    'GWS - UPSTREAM',
    'IP PEERING - NON R&E (PRIVATE)',
    'IP PEERING - NON R&E (PUBLIC)',
    'IP PEERING - R&E',
    'IP TRUNK',
    'L2SERVICES',
    'L3-VPN',
    'MD-VPN (INTERNAL)',
    'MD-VPN (NATIVE)',
    'MD-VPN (PROXY)',
    'POP LAN LINK',
    'SERVER LINK'
}


class CustomerType(Enum):
    COMMERCIAL_PEER = 1
    CONNECTIVITY_SUPPLIER = 2
    EU_NREN = 3
    HOUSING_SUPPLIER = 4
    IX_SUPPLIER = 5
    OTHER = 6
    R_AND_E_PEER = 7
    UNKNOWN = 'UNKNOWN'


class InventoryStatus(Enum):
    PLANNED = 1
    READY_FOR_SERVICE = 2
    IN_SERVICE = 3
    MIGRATION = 4
    OUT_OF_SERVICE = 6
    READY_FOR_CEASURE = 7


class RelateType(Enum):
    VENDOR = 1
    CONTACT = 2
    GROUP = 3
    CONTRACT = 4
    CUSTOMER = 5
    SITE = 6


class IMSError(Exception):
    pass


class IMS(object):

    PERMITTED_RECONNECT_ATTEMPTS = 3
    LOGIN_PATH = '/login'
    LOGOUT_PATH = '/login/logout'
    HEARTBEAT_PATH = '/login/heartbeat'
    IMS_PATH = '/ims'

    def __init__(self, base_url, username, password, verify_ssl=False):
        self.base_url = base_url
        self.username = username
        self.password = password
        self.reconnect_attempts = 0
        self.session = requests.session()
        self.session.verify = verify_ssl

    @property
    def bearer_token(self):
        try:
            return self.session.headers['Authorization'].replace('Bearer ', '')
        except KeyError:
            return None

    @bearer_token.setter
    def bearer_token(self, value):
        if value is None:
            self.session.headers.pop('Authorization', None)
        else:
            self.session.headers.update({'Authorization': f'Bearer {value}'})

    def __enter__(self):
        self._init_bearer_token()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.logout()

    def __del__(self):
        self.logout()

    def _is_heartbeat_ok(self):
        if not self.bearer_token:
            return False

        for i in range(IMS.PERMITTED_RECONNECT_ATTEMPTS):
            try:
                response = self.session.get(self.base_url + self.HEARTBEAT_PATH)
                return response.status_code == requests.codes.ok
            except HTTPError as e:
                logger.error(f'Error checking heartbeat (attempt {i + 1}): {e}')
        raise IMSError("Failed to get heartbeat")

    def _init_bearer_token(self):
        if not self.bearer_token:
            self.reconnect_attempts = 0
        elif self._is_heartbeat_ok():
            return
        else:
            self.reconnect_attempts += 1
            if self.reconnect_attempts > self.PERMITTED_RECONNECT_ATTEMPTS:
                raise IMSError('Too many reconnection attempts made')

        logger.debug(f'Logging in - Username: {self.username}'
                     f' - URL: {self.base_url + self.LOGIN_PATH}')
        response = self.session.post(
            self.base_url + self.LOGIN_PATH,
            auth=(self.username, self.password))
        response.raise_for_status()
        self.bearer_token = response.text

    def logout(self):
        if self.bearer_token is None:
            return
        logout_url = self.base_url + self.LOGOUT_PATH
        response = self.session.get(logout_url)
        try:
            response.raise_for_status()
        except HTTPError as e:
            logger.error(f'Error logging out: {e}')
        self.bearer_token = None

    def clear_dynamic_context_cache(self):
        if not self.bearer_token:
            self._init_bearer_token()
        while True:
            logger.info('Clearing Dynamic Context Cache')
            response = self.session.put(
                f'{self.base_url + IMS.IMS_PATH}/ClearDynamicContextCache',
            )
            if response.status_code == 401:
                self._init_bearer_token()
                continue
            response.raise_for_status()
            break

    def _get_ims_data(
            self,
            url,
            params=None,
            navigation_properties=None):
        url = f'{self.base_url + IMS.IMS_PATH}/{url}'
        if navigation_properties:
            params = params if params else {}
            params['navigationproperty'] = navigation_properties if isinstance(
                navigation_properties, int) else sum(navigation_properties)

        if not self.bearer_token:
            self._init_bearer_token()

        def _is_invalid_login_state(response_):
            if response_.status_code == requests.codes.unauthorized:
                return True

            if response_.status_code in (requests.codes.ok,
                                         requests.codes.not_found):

                if NO_FILTERED_RESULTS_MESSAGE in response_.text.lower():
                    return False
                try:
                    r = _convert_keys(response_.json())
                except Exception as e:
                    t = response_.text
                    if len(t) > 100:
                        message_text = f'{t[:50]} ... {t[-50:]}'
                    else:
                        message_text = t
                    logger.debug(f'unexpected response: {message_text}\n{e}'
                                 '\nre-raising')
                    raise e
                if r and 'haserrors' in r and r['haserrors']:
                    for e in r['Errors']:
                        if e['errormessage'] and \
                                'guid expired' in e['errormessage'].lower():
                            return True

            return False

        def _convert_keys(source):
            if isinstance(source, list):
                return [_convert_keys(x) for x in source]
            elif isinstance(source, dict):
                new = {}
                for k, v in source.items():
                    if isinstance(v, (dict, list)):
                        v = _convert_keys(v)
                    new[k.lower()] = v
                return new
            return source

        while True:
            response = self.session.get(url, params=params)
            if _is_invalid_login_state(response):
                self._init_bearer_token()
            else:
                response.raise_for_status()
                orig = response.json()
                return_value = _convert_keys(orig)
                return return_value

    @log_entry_and_exit
    def get_entity_by_id(
            self,
            entity_type,
            entity_id,
            navigation_properties=None):

        url = f'{entity_type}/{entity_id}'
        return \
            self._get_ims_data(url, None, navigation_properties)

    @log_entry_and_exit
    def get_entity_by_name(
            self,
            entity,
            name,
            navigation_properties=None):
        url = f'{entity}/byname/"{name}"'
        return self._get_ims_data(url, None, navigation_properties)

    @log_entry_and_exit
    def get_filtered_entities(
            self,
            entity,
            filter_string,
            navigation_properties=None,
            step_count=50):
        url = f'{entity}/filtered/{filter_string}'
        yield from self.get_entities(
            url,
            navigation_properties,
            step_count
        )

    def get_entities(
            self,
            url,
            navigation_properties=None,
            step_count=50):
        more_to_come = True
        start_entity = 0
        gateway_error_count = 0
        while more_to_come:
            params = {
                'paginatorStartElement': start_entity,
                'paginatorNumberOfElements': step_count
            }
            try:
                more_to_come = False
                entities = self._get_ims_data(
                    url,
                    params,
                    navigation_properties)
            except HTTPError as e:
                r = e.response
                if r.status_code == requests.codes.not_found \
                        and NO_FILTERED_RESULTS_MESSAGE in r.text.lower():
                    entities = None
                elif r.status_code == 504:
                    gateway_error_count += 1
                    logger.debug(
                        f'GATEWAY TIME-OUT for {url}'
                        f'  --  COUNT: {gateway_error_count}')
                    if gateway_error_count > 4:
                        raise e
                    time.sleep(5)
                    logger.debug('WAKING UP')
                    more_to_come = True
                    continue
                else:
                    raise e
            if entities:
                more_to_come = \
                    len(entities) >= step_count
                start_entity += step_count
                yield from entities

    @log_entry_and_exit
    def get_all_entities(
            self,
            entity,
            navigation_properties=None,
            step_count=50
    ):
        url = f'{entity}/all'
        yield from self.get_entities(
            url,
            navigation_properties,
            step_count
        )
