import json

import requests
from requests import HTTPError
import pytest
from requests.models import Response

from inventory_provider.db.ims import IMS, IMSError


class PatchedIMS(IMS):

    def __del__(self):
        # patched to stop call to logout
        pass


BASE_URL = "http://dummy_base"
USERNAME = "dummy_username"
PASSWORD = "dummy_password"


@pytest.fixture
def ims_instance(mocker):
    ims_instance = PatchedIMS(BASE_URL, USERNAME, PASSWORD)
    mock_post = mocker.patch.object(ims_instance.session, 'post')
    mock_post.return_value.status_code = 200
    mock_post.return_value.text = 'dummy_bt'
    return ims_instance


def test_init_bearer_token(ims_instance):

    ims_instance._init_bearer_token()
    assert ims_instance.bearer_token == 'dummy_bt'
    assert ims_instance.reconnect_attempts == 0


def test_heartbeat(mocker, ims_instance):

    # there should be no bearer token set
    assert ims_instance._is_heartbeat_ok() is False

    # set the bearer token
    ims_instance.bearer_token = 'dummy_bt'

    mocked_get = mocker.patch.object(ims_instance.session, 'get')

    mocked_get.return_value.status_code = 200
    assert ims_instance._is_heartbeat_ok()

    mocked_get.return_value.status_code = 401
    assert ims_instance._is_heartbeat_ok() is False

    dummy_exception = HTTPError('http://dummy_base/heartbeat', 500, 'Dummy', None, None)
    responses = [dummy_exception] * IMS.PERMITTED_RECONNECT_ATTEMPTS
    mocked_get.side_effect = responses
    with pytest.raises(IMSError) as exception_info:
        ims_instance._is_heartbeat_ok()
        assert exception_info.value.message == 'Dummy'

    r = Response()
    r.status_code = 200
    responses[-1] = r
    mocked_get.side_effect = responses
    assert ims_instance._is_heartbeat_ok()

    r.status_code = 401
    responses[-1] = r
    mocked_get.side_effect = responses
    assert ims_instance._is_heartbeat_ok() is False


def test_get_entity_by_id(mocker, ims_instance):
    entity = 'dummy_entity_type'
    entity_id = '12345'
    expected_entity = {'id': entity_id, 'name': 'Test Entity'}

    mocked_get = mocker.patch.object(ims_instance.session, 'get')
    mocked_get.return_value.status_code = 200
    mocked_get.return_value.json.return_value = expected_entity

    result = ims_instance.get_entity_by_id(entity, entity_id)

    assert result == expected_entity


def test_get_entity_by_name(mocker, ims_instance):
    entity = 'dummy_entity_tyoe'
    entity_name = 'dummy_entity_name'
    expected_entity = {'id': '1234', 'name': entity_name}

    mocked_get = mocker.patch.object(ims_instance.session, 'get')
    mocked_get.return_value.status_code = 200
    mocked_get.return_value.json.return_value = expected_entity

    result = ims_instance.get_entity_by_name(entity, entity_name)

    assert result == expected_entity


def test_get_filtered_entities(mocker, ims_instance):
    entity = 'dummy_entity_tyoe'
    entity_filter = 'dummy_param=dummy value'
    expected_result = [{'id': '1234', 'name': 'entity_name'}]

    mocked_get = mocker.patch.object(ims_instance.session, 'get')
    mocked_get.return_value.status_code = 200
    mocked_get.return_value.json.return_value = expected_result

    result = ims_instance.get_filtered_entities(entity, entity_filter)

    assert list(result) == expected_result


def test_get_entities(mocker, ims_instance):
    entity = 'dummy_entity_type'
    expected_result = [{'id': '1234', 'name': 'entity_name'}]

    mocked_get = mocker.patch.object(ims_instance.session, 'get')
    mocked_get.return_value.status_code = 200
    mocked_get.return_value.json.return_value = expected_result

    result = ims_instance.get_entities(entity)

    assert list(result) == expected_result


@pytest.mark.parametrize('step_count', [1, 3, 4, 9, 10])
def test_get_paged_entities(step_count, ims_instance, mocker):
    entity = 'dummy_entity_type'
    expected_result = [
        {'id': '12341', 'name': 'entity_name_1'},
        {'id': '12342', 'name': 'entity_name_2'},
        {'id': '12343', 'name': 'entity_name_3'},
        {'id': '12344', 'name': 'entity_name_4'},
        {'id': '12345', 'name': 'entity_name_5'},
        {'id': '12346', 'name': 'entity_name_6'},
        {'id': '12347', 'name': 'entity_name_7'},
        {'id': '12348', 'name': 'entity_name_8'},
        {'id': '12349', 'name': 'entity_name_9'},
    ]
    mocked_get = mocker.patch.object(ims_instance.session, 'get')
    mocked_responses = []

    for i in range(0, len(expected_result) + 1, step_count):
        resp = Response()
        resp.status_code = 200
        if i >= len(expected_result):
            resp_content = {}
        else:
            resp_content = expected_result[i:i + step_count]
        resp._content = str.encode(json.dumps(resp_content))
        mocked_responses.append(resp)
    mocked_get.side_effect = mocked_responses

    result = ims_instance.get_all_entities(entity, step_count=step_count)
    assert list(result) == expected_result


def test_context_manager(ims_instance, mocker):
    mocked_get = mocker.patch.object(ims_instance.session, 'get')
    mocked_get.return_value.status_code = 200
    mocked_get.return_value.json.return_value = []

    with ims_instance as ims:
        list(ims_instance.get_all_entities('dummy'))
        assert ims.bearer_token == 'dummy_bt'

    assert ims_instance.bearer_token is None


def test_navigation_properties(ims_instance, mocker):
    mocked_get = mocker.patch.object(ims_instance.session, 'get')

    ims_instance.get_entity_by_id('Node', 1234, navigation_properties=[1, 2, 3])
    mocked_get.assert_called_with(
        f'{BASE_URL}{IMS.IMS_PATH}/Node/1234', params={'navigationproperty': 6})


def test_get_entities_gateway_error(ims_instance, mocker):
    mocked_get = mocker.patch.object(ims_instance.session, 'get')
    mocker.patch('inventory_provider.db.ims.time')

    error_message = 'Dummy Gateway Error'
    r = Response()
    r.status_code = requests.codes.gateway_timeout
    r._content = str.encode(error_message)
    mocked_get.side_effect = [r] * 5
    with pytest.raises(HTTPError) as exception_info:
        list(ims_instance.get_entities('dummy'))
        assert exception_info.value.message == error_message
        assert mocked_get.call_count == 4
