import json
import os

import jsonschema

import inventory_provider
from inventory_provider.db.ims import InventoryStatus
from inventory_provider.db.ims_data import lookup_lg_routers, \
    get_node_locations, IMS_OPSDB_STATUS_MAP, \
    get_port_id_services, get_port_details, \
    get_circuit_hierarchy, get_ids_and_sids, NODE_LOCATION_SCHEMA, \
    get_flexils_by_circuitid, SITE_LOCATION_SCHEMA, get_site_locations, \
    get_equipment_details


def _json_test_data(filename):
    abs_filename = os.path.join(
        os.path.dirname(__file__),
        'data',
        filename)
    with open(abs_filename) as data:
        return json.load(data)


def test_get_circuit_hierarchy(mocker):

    ds = inventory_provider.db.ims.IMS(
        'http://dummy_base', 'dummy_username', 'dummy_password')
    se_data = _json_test_data('ims_circuit_hierarchy_data.json')
    mocker.patch.object(
        inventory_provider.db.ims.IMS,
        'get_filtered_entities',
        side_effect=[se_data, [
            {'selection': 'IP PEERING - R&E'}
        ]]
    )
    res = list(get_circuit_hierarchy(ds))
    assert ds.get_filtered_entities.call_count == 2
    predicted = [
        {
            'id': 661591,
            'name': 'UK_ORIENT',
            'status': 'operational',
            'product': 'IP PEERING - R&E',
            'speed': 'ETHS',
            'project': 'ORIENTPLUS',
            'circuit-type': 'service',
            'sub-circuits': [],
            'carrier-circuits': [660461],
            'customerid': 57773
        },
        {
            'id': 660461,
            'name': 'V-LAN_200_MX1.LON.UK_XE-0/0/1_',
            'status': 'operational',
            'product': 'ETHERNET',
            'speed': 'V-LAN',
            'project': 'ORIENTPLUS',
            'circuit-type': 'circuit',
            'sub-circuits': [661591],
            'carrier-circuits': [668866],
            'customerid': 57773
        }
    ]
    assert res == predicted


def test_get_port_details(mocker):

    def _se(entity, y, step_count):
        return _json_test_data(f'ims_{entity}_details_data.json')

    mocker.patch.object(
        inventory_provider.db.ims.IMS,
        'get_all_entities',
        side_effect=_se
    )

    def _ports(type):
        def func(*args, **kwargs):
            _data = _json_test_data(f'ims_{type}_ports_sids.json')

            return {int(k): v for k, v in _data.items()}
        return func

    _external = mocker.patch.object(
        inventory_provider.db.ims_data,
        'get_port_sids',
        side_effect=_ports('external')
    )

    _internal = mocker.patch.object(
        inventory_provider.db.ims_data,
        'get_internal_port_sids',
        side_effect=_ports('internal')
    )

    ds = inventory_provider.db.ims.IMS(
        'dummy_base', 'dummy_username', 'dummy_password')
    res = list(get_port_details(ds))
    assert ds.get_all_entities.call_count == 2
    assert _external.call_count == 1
    assert _internal.call_count == 1

    predicted = [
        {
            'port_id': 6417059,
            'equipment_name': 'RT1.TAL.EE',
            'interface_name': 'XE-0/1/7'
        },
        {
            'port_id': 6417563,
            'equipment_name': 'SW1.CH.OFFICE.GEANT.NET',
            'interface_name': 'XE-0/2/0'
        },
        {
            'port_id': 6420287,
            'equipment_name': 'GEN01-DTNX10-1',
            'interface_name': '1-A-1-L1'
        },
        {
            'port_id': 222451,
            'equipment_name': 'QFX.FRA.DE',
            'interface_name': 'AE0',
            'sid': 'GA-01672'
        },
        {
            'port_id': 222452,
            'equipment_name': 'QFX.FRA.DE',
            'interface_name': 'AE1',
            'sid': 'GA-01673'
        },
        {
            'port_id': 225742,
            'equipment_name': 'MX1.MAD.ES',
            'interface_name': 'AE3.103',
            'sid': 'GS-00121'
        }
    ]
    assert res == predicted


def test_get_port_id_services(mocker):
    d = _json_test_data('ims_port_id_services_data.json')

    mocker.patch.object(
        inventory_provider.db.ims.IMS,
        'get_all_entities',
        side_effect=[
            [
                {'id': 57658, 'name': 'ORG A'},
                {'id': 57664, 'name': 'ORG B'},
                {'id': 57640, 'name': 'ORG C'},
                {'id': 57744, 'name': 'ETH'},
            ],
            [
                {'id': 3800, 'name': 'PRODUCT A'},
                {'id': 3804, 'name': 'GEANT IP'},
                {'id': 3810, 'name': 'GEANT PEERING'},
                {'id': 3677, 'name': 'ETHERNET'},
            ]]
    )
    se = [
        [
            {'selection': 'GEANT IP'},
            {'selection': 'GEANT PEERING'},
            {'selection': 'PRODUCT A'}
        ]
    ]
    se.extend(d)
    mocker.patch.object(
        inventory_provider.db.ims.IMS,
        'get_filtered_entities',
        side_effect=se
    )

    ds = inventory_provider.db.ims.IMS(
        'http://dummy_base', 'dummy_username', 'dummy_password')
    res = list(get_port_id_services(ds))

    assert ds.get_all_entities.call_count == 2
    assert ds.get_filtered_entities.call_count == 4
    predicted = [
        {
            'id': 663060,
            'name': 'ORG_A_AP2',
            'status': 'operational',
            'circuit_type': 'service',
            'service_type': 'GEANT IP',
            'project': 'ORG A',
            'customer': 'ORG A',
            'port_a_id': 224507,
            'customerid': 57658,
            'port_type': 'internal'
        },
        {
            'id': 663104,
            'name': 'ORG_B_AP2_IAS',
            'status': 'operational',
            'circuit_type': 'service',
            'service_type': 'GEANT PEERING',
            'project': 'ORG B',
            'customer': 'ORG B',
            'port_a_id': 224464,
            'customerid': 57664,
            'port_type': 'internal'
        },
        {
            'id': 679324,
            'name': 'AMSTERDAM-LONDON-100GBE-001(ETH)',
            'status': 'operational',
            'circuit_type': 'circuit',
            'service_type': 'ETHERNET',
            'project': 'ETH',
            'customer': 'ETH',
            'port_a_id': 6423107,
            'port_b_id': 6419340,
            'customerid': 57744,
            'port_type': 'ports'
        },
        {
            'id': 679324,
            'name': 'AMSTERDAM-LONDON-100GBE-001(ETH)',
            'status': 'operational',
            'circuit_type': 'circuit',
            'service_type': 'ETHERNET',
            'project': 'ETH',
            'customer': 'ETH',
            'port_a_id': 6419340,
            'port_b_id': 6423107,
            'customerid': 57744,
            'port_type': 'ports'
        },
        {
            'id': 679324,
            'name': 'AMSTERDAM-LONDON-100GBE-001(ETH)',
            'status': 'operational',
            'circuit_type': 'circuit',
            'service_type': 'ETHERNET',
            'project': 'ETH',
            'customer': 'ETH',
            'port_a_id': 6423111,
            'customerid': 57744,
            'port_type': 'ports'
        },
        {
            'id': 702560,
            'name': 'VMPORT Circuit 1',
            'status': 'operational',
            'circuit_type': 'service',
            'service_type': 'PRODUCT A',
            'project': 'ORG C',
            'customer': 'ORG C',
            'port_a_id': 6419453,
            'customerid': 57640,
            'port_type': 'port relate'
        }
    ]
    assert res == predicted


def test_lookup_lg_routers(mocker):

    ims = mocker.patch('inventory_provider.db.ims.IMS')
    ims.return_value.get_filtered_entities.return_value \
        = _json_test_data('ims_lg_data.json')
    ims.return_value.get_entity_by_id.return_value = {
        'name': 'pop name',
        'longitude': 'long',
        'latitude': 'lat',
        'city': {
            'name': 'city name',
            'country': {
                'name': 'country name',
                'abbreviation': 'country code'
            }
        },
        'sitealiases': [
            {
                'aliasname': 'abbr'
            }
        ]
    }

    ds = inventory_provider.db.ims.IMS(
        'dummy_base', 'dummy_username', 'dummy_password')

    res = list(lookup_lg_routers(ds))

    ds.get_filtered_entities.assert_called_once_with(
        'EquipmentDefinition',
        'Name like MX',
        inventory_provider.db.ims.EQUIP_DEF_PROPERTIES['Nodes'])

    assert ds.get_entity_by_id.call_count == 35
    assert len(res) == 35

    pop = {
        'name': 'pop name',
        'city': 'city name',
        'country': 'country name',
        'country code': 'country code',
        'abbreviation': 'abbr',
        'longitude': 'long',
        'latitude': 'lat'
    }
    assert res[0] == {
        'equipment name': 'MX3.LAB.OFFICE.GEANT.NET',
        'type': 'CORE',
        'pop': pop
    }


def test_get_node_location(mocker):
    ims = mocker.patch('inventory_provider.db.ims.IMS')
    resp_data = _json_test_data('ims_nodes_data.json')
    ims.return_value.get_all_entities.return_value = resp_data

    ds = inventory_provider.db.ims.IMS(
        'dummy_base', 'dummy_username', 'dummy_password')
    res = list(get_node_locations(ds))
    for name, node in res:
        assert isinstance(name, str)
        jsonschema.validate(node, NODE_LOCATION_SCHEMA)

    assert len(res) == 36
    assert res[0] == ('LON3_CX_01', {
        'equipment-name': 'LON3_CX_01',
        'status': IMS_OPSDB_STATUS_MAP[InventoryStatus.IN_SERVICE],
        'pop': {
            'name': 'LONDON 3 POWERGATE',
            'city': 'LONDON',
            'country': 'UNITED KINGDOM',
            'abbreviation': 'LON3',
            'longitude': -0.257712,
            'latitude': 51.5308142,
        }
    })


def test_get_site_location(mocker):
    ims = mocker.patch('inventory_provider.db.ims.IMS')
    resp_data = _json_test_data('ims_nodes_data.json')
    ims.return_value.get_all_entities.return_value = resp_data

    ds = inventory_provider.db.ims.IMS(
        'dummy_base', 'dummy_username', 'dummy_password')
    res = list(get_site_locations(ds))
    for name, site in res:
        assert isinstance(name, str)
        jsonschema.validate(site, SITE_LOCATION_SCHEMA)

    assert len(res) == 2
    assert res[0] == ('LONDON 3 POWERGATE',
                      {'abbreviation': 'LON3', 'city': 'LONDON', 'country': 'UNITED KINGDOM', 'latitude': 51.5308142,
                       'longitude': -0.257712, 'name': 'LONDON 3 POWERGATE'})


def test_get_circuit_ids_and_sids(mocker):
    ims = mocker.patch('inventory_provider.db.ims.IMS')
    ims.return_value.get_filtered_entities.return_value = \
        [
            {
                'id': 999999,
                'extrafieldid': 3209,
                'objectid': 111111,
                'value': 'SID-01',
                'extrafield': None,
                'links': None,
                'extrafieldvalueobjectinfo': None,
                'rowversion': '2021-12-10T19:40:51',
                'errors': None,
                'haserrors': False
            }, {
                'id': 999998,
                'extrafieldid': 3209,
                'objectid': 111112,
                'value': 'SID-02',
                'extrafield': None,
                'links': None,
                'extrafieldvalueobjectinfo': None,
                'rowversion': '2021-12-10T19:40:51',
                'errors': None,
                'haserrors': False
            }, {
                'id': 999997,
                'extrafieldid': 3209,
                'objectid': 111113,
                'value': 'SID-03',
                'extrafield': None,
                'links': None,
                'extrafieldvalueobjectinfo': None,
                'rowversion': '2021-12-10T19:40:51',
                'errors': None,
                'haserrors': False
            },
        ]
    expected_response = [
        (111111, 'SID-01'),
        (111112, 'SID-02'),
        (111113, 'SID-03')
    ]

    ds = inventory_provider.db.ims.IMS(
        'dummy_base', 'dummy_username', 'dummy_password')
    res = list(get_ids_and_sids(ds))
    assert res == expected_response


def test_get_flexils_by_circuit_id(mocker):
    ims = mocker.patch('inventory_provider.db.ims.IMS')
    ims.return_value.get_all_entities.return_value = \
        [
            {
                "$id": "1630",
                "circuitid": 702203,
                "circuitname": "GEANT LAB-GEANT LAB-OCH-003",
                "id": 1,
                "nodename": "CAM01-MTC6-3",
                "node_id": 130498,
                "port": "A-1-S1",
                "port_ref": "1-A-1-S1-1",
                "shelfname": "1-MTC-6",
                "rowVersion": "0001-01-01T00:00:00",
                "circuit": None,
                "node_": None,
                "errors": None,
                "haserrors": False,
                "customproperties": {}
            },
            {
                "$id": "1631",
                "circuitid": 702203,
                "circuitname": "GEANT LAB-GEANT LAB-OCH-003",
                "id": 2,
                "nodename": "CAM01-MTC6-3",
                "node_id": 130498,
                "port": "A-1-L1",
                "port_ref": "1-A-1-L1-1",
                "shelfname": "1-MTC-6",
                "rowversion": "0001-01-01T00:00:00",
                "circuit": None,
                "node_": None,
                "errors": None,
                "haserrors": False,
                "customproperties": {}
            }
        ]
    expected_result = {
        702203: [
            {
                'node_name': 'CAM01-MTC6-3',
                'full_port_name': '1-A-1-S1-1',
                'key': 'CAM01-MTC6-3:1-A-1-S1-1'
            },
            {
                'node_name': 'CAM01-MTC6-3',
                'full_port_name': '1-A-1-L1-1',
                'key': 'CAM01-MTC6-3:1-A-1-L1-1'
            }
        ]
    }

    ds = inventory_provider.db.ims.IMS(
        'dummy_base', 'dummy_username', 'dummy_password')
    res = get_flexils_by_circuitid(ds)
    assert res == expected_result


def test_get_equipment_details(mocker):
    ims_data = [
        {
            'id': 1001,
            'name': 'Dummy ED 1',
            'vendorId': 9001,
            'vendor': {
                'id': 9001,
                'name': 'Dummy Vendor 1',
                    'typeId': 1,
                    'slaOnSiteId': None,
                    'slaMainTimeToRepairId': None,
            },
            'nodes': [
                {
                    'id': 7001,
                    'name': 'Dummy Node 1',
                    'equipmentDefinitionId': 1001,
                    'inventoryStatusId': 3,
                    'siteId': 6001,
                    'ringId': 5001,
                },
                {
                    'id': 7002,
                    'name': 'Dummy Node 2',
                    'equipmentDefinitionId': 1001,
                    'inventoryStatusId': 2,
                    'siteId': 6001,
                    'ringId': 5002,
                }
            ]
        },
        {
            'id': 1002,
            'name': 'Dummy ED 2',
            'vendorId': 9002,
            'vendor': {
                'id': 9002,
                'name': 'Dummy Vendor 2',
                    'typeId': 1,
                    'slaOnSiteId': None,
                    'slaMainTimeToRepairId': None,
            },
            'nodes': [
                {
                    'id': 7003,
                    'name': 'Dummy Node 3',
                    'equipmentDefinitionId': 1002,
                    'inventoryStatusId': 3,
                    'siteId': 6001,
                    'ringId': 5001,
                }
            ]
        },
        {
            'id': 1003,
            'name': 'Dummy ED 3',
            'vendorId': 9002,
            'vendor': {
                'id': 9002,
                'name': 'Dummy Vendor 2',
                    'typeId': 1,
                    'slaOnSiteId': None,
                    'slaMainTimeToRepairId': None,
            },
            'nodes': []
        }
    ]
    ims = mocker.patch('inventory_provider.db.ims.IMS')
    ims.return_value.get_all_entities.return_value = ims_data

    ds = inventory_provider.db.ims.IMS('xxx', 'xxx', 'xxx')
    equipment_details = list(get_equipment_details(ds))
    expected_result = [
        {'name': 'Dummy Node 1', 'model': 'Dummy ED 1', 'vendor': 'Dummy Vendor 1'},
        {'name': 'Dummy Node 2', 'model': 'Dummy ED 1', 'vendor': 'Dummy Vendor 1'},
        {'name': 'Dummy Node 3', 'model': 'Dummy ED 2', 'vendor': 'Dummy Vendor 2'}
    ]
    assert equipment_details == expected_result
